import time

import colorsys

from adapter import OpenDMXAdapter
from fixtures.cameo.rootpar6 import RootPar6

controller = OpenDMXAdapter("ftdi://ftdi:232:BG00DND8/1")
controller.start()
controller.blackout()

light = RootPar6(0)
controller.add_fixture(light)


for i in range(361):
    r, g, b = colorsys.hsv_to_rgb(i / 360, 1, 1)
    print(f"\r{r}/{g}/{b}", end="")
    light.set_rgb(int(r * 255), int(g * 255), int(b * 255))
    time.sleep(0.01)

print("\nDone!")



from ..basefixture import BaseFixture, ColorFixture, StroboFixture


class RootPar6(BaseFixture, ColorFixture, StroboFixture):
    def __init__(self, raw_channel: int | None = None):
        """
        8ch mode
        0 - dimmer
        1 - strobe
        2 - red
        3 - green
        4 - blue
        5 - white
        6 - amber
        7 - uv
        """
        super().__init__(8, raw_channel)
        self._initialize_color_channels(2, 3, 4, 0)
        self._initialize_strobo_channels(1)

    def set_white(self, value: int):
        self.set_value(5, value)

    def set_amber(self, value: int):
        self.set_value(6, value)

    def set_uv(self, value: int):
        self.set_value(7, value)



from ..basefixture import BaseFixture, ColorFixture, MovingHeadFixture, StroboFixture


class TMHX4(BaseFixture, ColorFixture, MovingHeadFixture, StroboFixture):
    def __init__(self, raw_channel: int | None = None):
        """
        16ch mode
        ...

        24ch mode
        0 - Pan Clockwise
        1 - Pan fine
        2 - Tilt
        3 - Tilt fine
        4 - Moving Speed (1-212: less speed, 213-255: nothing)
        5 - Lens Zoom
        6 - Intensity
        7 - Strobo (0-3: None, 4-95: normal, 96-176: random, 177-255: "thunder"-strobe)

        8 - Red     Inner
        9 - Green   Inner
        10 - Blue   Inner
        11 - White  Inner

        12 - Red    Outer (Top-Start)
        13 - Green  Outer (Top-Start)
        14 - Blue   Outer (Top-Start)
        15 - White  Outer (Top-Start)

        16 - Red    Outer (Non-Top-Start)
        17 - Green  Outer (Non-Top-Start)
        18 - Blue   Outer (Non-Top-Start)
        19 - White  Outer (Non-Top-Start)

        20 - Color Preset and Macro
        21 - Macro Speed
        22 - Pattern (0-49: Channel 1-20, 50-99: Presets/Macros, 100-149: Internal P1, 150-199: Internal P2, 200-255: Music)
        23 - Reset (251-255)
        """
        super().__init__(24, raw_channel)
        self._initialize_color_channels(8, 9, 10, 6)
        self._initialize_moving_head_channels(0, 1, 2, 3, 4)
        self._initialize_strobo_channels(7)

    def set_outer_color(self, r: int, g: int, b: int, w: int, top_start: bool = True):
        self.set_value(12 + 4 if top_start else 0, r)
        self.set_value(13 + 4 if top_start else 0, g)
        self.set_value(14 + 4 if top_start else 0, b)
        self.set_value(15 + 4 if top_start else 0, w)

    def set_lens_zoom(self, value: int):
        self.set_value(5, value)

    def set_pattern(self, value: int):
        self.set_value(22, value)

    def set_reset(self, should_reset: bool):
        self.set_value(23, 255 if should_reset else 0)

    def set_color_macro(self, value: int):
        self.set_value(20, value)

    def set_macro_speed(self, value: int):
        self.set_value(21, value)



import sys
import time
import threading

import pyftdi.serialext
from serial import serialutil

from . import fixtures


class OpenDMXAdapter:
    def __init__(self, serial_port: str):
        """
        Initializes the OpenDMXAdapter with the specified serial port.
        Use utils.helper.list_devices() to find the correct serial uri.

        :param serial_port: Something like "ftdi://ftdi:232:BG00DND8/1"
        """
        self.is_connected = False

        try:
            self.serial = pyftdi.serialext.serial_for_url(serial_port, baudrate=250000, stopbits=2)
        except ValueError as e:
            print("Malformed serialPort url: " + serial_port)
            print(e)
            sys.exit(0)
        except serialutil.SerialException as e:
            print("Error: could not open Serial Port")
            print(e)
            sys.exit(0)

        self.is_connected = True
        self.dmx_data = [bytes([0])] * 513
        self.dmx_data[0] = bytes([0])
        self.display_thread = threading.Thread(target=self._display_universe)
        self.fixtures = []
        self.channel_index = 0

    def add_fixture(self, fixture: 'fixtures.basefixture.BaseFixture') -> None:
        """
        Adds a fixture to the DMX adapter which will be initialized with the next available channel index.
        :param fixture: An instance of a fixture that inherits from BaseFixture.
        :return: None
        """
        fixture.adapter = self
        self.channel_index += fixture.initialize(self.channel_index)

        if self.channel_index > len(self.dmx_data) - 1:
            raise RuntimeError("Channel index out of range")

        self.fixtures.append(fixture)

    def add_fixtures(self, *fixture_list: 'fixtures.basefixture.BaseFixture') -> None:
        """
        Adds multiple fixtures to the DMX adapter.
        :param fixture_list: A variable number of fixture instances that inherit from BaseFixture.
        :return: None
        """
        for fixture in fixture_list:
            self.add_fixture(fixture)

    def set_channel(self, channel, intensity) -> None:
        """
        Sets the intensity for a specific DMX channel.

        :param channel: The DMX channel number (0-512).
        :param intensity: The intensity value (0-255) to set for the channel.
        :return: None
        """
        channel = max(0, min(512, channel))
        intensity = max(0, min(255, intensity))
        self.dmx_data[channel + 1] = bytes([intensity])

    def blackout(self) -> None:
        """
        Sets all DMX channels to zero intensity, effectively blacking out the output.
        :return: None
        """
        for i in range(1, 512, 1):
            self.dmx_data[i] = bytes([0])

    def start(self) -> None:
        """
        Starts the display thread that continuously renders the DMX universe.
        :return: None
        """
        self.display_thread.start()

    def close(self):
        """
        Closes the serial connection and stops the display thread.
        :return: None
        """
        self.is_connected = False
        self.display_thread.join()
        self.serial.close()

    def _display_universe(self):
        while self.is_connected:
            self._render()
            time.sleep(8 / 1000.0)  # 40 Hz for Enttec Open DMX USB

    def _render(self):
        if not self.is_connected:
            return

        sdata = b''.join(self.dmx_data)
        self.serial.send_break(duration=0.001)
        self.serial.write(sdata)


# Guide to install driver: https://learn.adafruit.com/circuitpython-on-any-computer-with-ft232h/windows
if __name__ == '__main__':
    url = "ftdi://ftdi:232:BG00DND8/1"

    dmx = OpenDMXAdapter(url)
    dmx.start()

    dmx.set_channel(1, 200)
    dmx.set_channel(3, 100)

    print("Start fading...")

    for i in range(0, 255):
        print(f"\r{int((i / 255) * 100)}% done", end='')
        dmx.set_channel(4, i)
        time.sleep(0.01)

    print("\nFading done.")
    time.sleep(5)

    print("Blackout")
    dmx.blackout()

    time.sleep(1)
    dmx.close()




from ..adapter import OpenDMXAdapter


class _AbstractFixture:
    adapter: OpenDMXAdapter

    def initialize(self, channel_index: int):
        ...

    def set_value(self, index: int, value: int):
        ...


class BaseFixture(_AbstractFixture):
    def __init__(self,
                 channel_count: int,
                 raw_channel: int | None = None):
        self.raw_channel = raw_channel
        self.channel_count = channel_count
        self.auto_calculate_channel = self.raw_channel is None

    def initialize(self, channel_index: int):
        if self.auto_calculate_channel:
            self.raw_channel = channel_index

        # Black out everything
        for i in range(self.channel_count):
            self.set_value(i, 0)

        return self.channel_count

    def set_value(self, index: int, value: int):
        if not isinstance(index, int):
            raise TypeError('index must be an integer')

        if not isinstance(value, int):
            raise TypeError('value must be an int')

        if index < 0 or index >= self.channel_count:
            raise IndexError

        value = min(255, max(0, int(value)))
        self.adapter.set_channel(self.raw_channel + index, value)


class ColorFixture(_AbstractFixture):
    _redChannel: int | None = None
    _greenChannel: int | None = None
    _blueChannel: int | None = None
    _intensityChannel: int | None = None

    def _initialize_color_channels(self, red_channel: int, green_channel: int, blue_channel: int,
                                   intensity_channel: int | None):
        self._redChannel = red_channel
        self._greenChannel = green_channel
        self._blueChannel = blue_channel
        self._intensityChannel = intensity_channel

    def set_rgb(self, r: int, g: int, b: int):
        self.set_value(self._redChannel, r)
        self.set_value(self._greenChannel, g)
        self.set_value(self._blueChannel, b)

    def set_intensity(self, intensity: int):
        self.set_value(self._intensityChannel, intensity)


class StroboFixture(_AbstractFixture):
    _stroboChannel: int | None = None

    def _initialize_strobo_channels(self, strobo_channel):
        self._stroboChannel = strobo_channel

    def set_strobo(self, value: int):
        self.set_value(self._stroboChannel, value)


class MovingHeadFixture(_AbstractFixture):
    _pan_channel: int | None = None
    _pan_fine_channel: int | None = None
    _tilt_channel: int | None = None
    _tilt_fine_channel: int | None = None
    _moving_speed_channel: int | None = None

    def _initialize_moving_head_channels(self, pan_channel: int, pan_fine_channel: int | None, tilt_channel: int,
                                         tilt_fine_channel: int | None, moving_speed_channel: int | None):
        self._pan_channel = pan_channel
        self._pan_fine_channel = pan_fine_channel
        self._tilt_channel = tilt_channel
        self._tilt_fine_channel = tilt_fine_channel
        self._moving_speed_channel = moving_speed_channel

    def set_pan(self, value: int):
        self.set_value(self._pan_channel, value)

    def set_pan_fine(self, value: int):
        self.set_value(self._pan_fine_channel, value)

    def set_tilt(self, value: int):
        self.set_value(self._tilt_channel, value)

    def set_tilt_fine(self, value: int):
        self.set_value(self._tilt_fine_channel, value)

    def set_moving_speed(self, speed: int):
        self.set_value(self._moving_speed_channel, speed)
